#!/usr/bin/env python3

# FIXME: proper __init__
# try:
#     from . import btle
#     from . import helpermaker
# except ImportError:
#     import btle
#     import helpermaker

# __all__: list[str] = ["btle", "helpermaker"]
